/*
Copyright © 2022 Sameer Rahmani <lxsameer@gnu.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
package cmd

import (
	"fmt"
	"os"
	"os/exec"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"
	"serene-lang.org/leavitt/pkg/core"
)

var cfgFile string
var debug bool
var state *core.State

// rootCmd represents the base command when called without any subcommands
var rootCmd = &cobra.Command{
	Use:   "leavitt",
	Short: "A source level TODO manager",
	Long: `A source level TODO manager which helps the devs to leave
TODOs and other important notes in form of comments on the source code`,
	// Uncomment the following line if your bare application
	// has an action associated with it:
	// Run: func(cmd *cobra.Command, args []string) { },
}

// Execute adds all child commands to the root command and sets flags appropriately.
// This is called by main.main(). It only needs to happen once to the rootCmd.
func Execute() {
	err := rootCmd.Execute()
	if err != nil {
		os.Exit(1)
	}
}

func init() {
	cobra.OnInitialize(initConfig)

	// Here you will define your flags and configuration settings.
	// Cobra supports persistent flags, which, if defined here,
	// will be global for your application.

	rootCmd.PersistentFlags().StringVar(&cfgFile, "config", "", "config file (default is $HOME/.leavitt.yaml)")

	// Cobra also supports local flags, which will only run
	// when this action is called directly.
	rootCmd.PersistentFlags().BoolP("debug", "d", false, "Turns On debug mode")
}

// initConfig reads in config file and ENV variables if set.
func initConfig() {
	var err error

	// go-git doesn't support this and life is too short to care
	// about that
	cmd := exec.Command("git", "rev-parse", "--show-toplevel")
	out, err := cmd.Output();
	cobra.CheckErr(err)

	path := string(out[:])
	cobra.CheckErr(err)

	if cfgFile != "" {
		// Use config file from the flag.
		viper.SetConfigFile(cfgFile)
	} else {
		// Find home directory.
		home, err := os.UserHomeDir()
		cobra.CheckErr(err)

		viper.AddConfigPath(home)
		viper.AddConfigPath(path)
		viper.SetConfigType("yaml")
		viper.SetConfigName(".leavitt")
	}

	viper.AutomaticEnv() // read in environment variables that match

	// If a config file is found, read it in.
	if err := viper.ReadInConfig(); err == nil {
		fmt.Fprintln(os.Stderr, "Using config file:", viper.ConfigFileUsed())
	}

	state, err = core.CreateState(path)
	cobra.CheckErr(err)
}
